(***********************************************************************
   Unit: DPMI.PAS
   Implementation of some DPMI-functions to Delphi
   J. Rathlev, Uni-Kiel, Apr. 1998

   The following functions are implemeted:
   - hook a protected mode call back into real mode interrupt
   - remove hooked interrupt
   - simulate real mode interrupt
   - enable virtual interrupt state
   - disable virtual interrupt state

   Using the first two functions allows to service a hardware interrupt
   from a Delphi program. The follwing example demonstrates the use of
   DPMI for this application:

  ***********************************************************************)

unit DPMI;

interface

uses SysUtils;

type
  {*** Delphi errors ***}
  EInterruptError = class(Exception);
  EDPMIError = class(EInterruptError);

  (* real mode register structure for DPMI-interface *)
  TRealModeRegs = record
    case Integer of
      0: (EDI, ESI, EBP, EXX, EBX, EDX, ECX, EAX: Longint;
          Flags, ES, DS, FS, GS, IP, CS, SP, SS: Word);
      1: (DI, DIH, SI, SIH, BP, BPH, XX, XXH: Word;
          case Integer of
          0: (BX, BXH, DX, DXH, CX, CXH, AX, AXH: Word);
          1: (BL, BH, BLH, BHH, DL, DH, DLH, DHH,
              CL, CH, CLH, CHH, AL, AH, ALH, AHH: Byte));
    end;

    StackArr = array [0..2] of word;

(* hook a protected mode call back into real mode interrupt *)
procedure InitRmInt (IntNr      : byte;
                     NewInt     : pointer;
                     var RmRegs : TRealModeRegs;
                     var OldInt : pointer);

(* remove hooked interrupt and restore old real mode interrupt *)
procedure RlsRmInt (IntNr  : byte;
                    OldInt : pointer);

(* simulate real mode interrupt *)
procedure RealModeInt (IntNr     : Byte;
                       var RmRegs: TRealModeRegs);

(* enable virtual interrupt state *)
procedure ClrInt;

(* disable virtual interrupt state *)
procedure SetInt;

implementation
const
  (* Fehlermelduungen *)
  DPMIError = 'Failed to execute DPMI interrupt';


{************************************************************************}
{***    DPMI-Interface                                                ***}
{************************************************************************}

(* ----------------------------------------------------------------------- *)
(* hook a protected mode call back into real mode interrupt
   IntNr:  number of interrupt to hook
   NewInt: address of Delphi interrupt routine
   RmRegs: address of real mode register structure
   OldInt: address of old real mode interrupt routine
           This address should be restored before closing the
           application by calling RlsRmInt *)
procedure InitRmInt (IntNr      : byte;
                     NewInt     : pointer;
                     var RmRegs : TRealModeRegs;
                     var OldInt : pointer);
begin
  asm
    push ds
    les  di,RmRegs          { Address to register struckture }
    lds  si,NewInt          { Address of prot. mode routine }
    mov  ax, 0303h
    int  31h
    pop  ds
    push dx                 { Real mode interrupt address }
    push cx

    mov  ax,$200            { save old interrupt vector }
    mov  bl,IntNr
    int  $31
    les  di,OldInt
    mov  es:[di],dx
    mov  es:[di+2],cx

    mov  ax,$201             { set new interruptvektor }
    mov  bl,IntNr
    pop  cx
    pop  dx
    int  $31
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* remove hooked interrupt and restore old real mode interrupt *)
procedure RlsRmInt (IntNr  : byte;
                    OldInt : pointer);
begin
  asm
    mov  ax,$201             { set old interruptvektor }
    mov  bl,IntNr
    mov  dx,word ptr OldInt
    mov  cx,word ptr OldInt+2
    int  $31
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* simulate real mode interrupt *)
procedure RealModeInt (IntNr      : Byte;
                       var RmRegs : TRealModeRegs);
var
  ErrorFlag: Boolean;
begin
  asm
    mov  ErrorFlag, 0       { assume success }
    mov  ax,0300h
    mov  bl,IntNr
    mov  bh,0
    mov  cx,0               { stack words to copy, assume zero }
    les  di,RmRegs          { es:di = register structure }
    int  31h
    jnc  @@End              { carry flag set on error }
    mov  ErrorFlag, 1
  @@End:
  end;
  if ErrorFlag then
    raise EDPMIError.Create(DPMIError);
end;

(* ----------------------------------------------------------------------- *)
(* enable virtual interrupt state *)
procedure ClrInt;
begin
  asm
    mov ax,$900
    int $31
    end;
  end;

(* ----------------------------------------------------------------------- *)
(* disable virtual interrupt state *)
procedure SetInt;
begin
  asm
    mov ax,$901
    int $31
    end;
  end;

end.